/*
By: Andrej Liptaj
Date: 17. 4. 2024
*/

import Jama.Matrix;
import java.util.Locale;
import java.security.SecureRandom;
import java.util.Random;

import static java.lang.Math.*;


public class Simple_fourier
	{
	static final double stat = 1E5;
	static final double eps = 1E-14;
	static final double[] cfs = {1.0, -1.0 / 3.0, 1.0 / 7.0, -1.0 / 11.0};
	final static double truth = 1.0;
	final static double width = 0.003;
	final static int nHistoBins = 50;

	static double fn(double x)
		{
		return cfs[0] + cfs[1] * cos(x) + cfs[2] * cos(2.0 * x) + cfs[3] * cos(3.0 * x);
		}

	static int binNb(double xMin, double xMax, int nBins, double x)
		{
		double dx = (xMax-xMin)/nBins;
		double delta = x-xMin;
		return (int)floor(delta/dx);
		}

	static double[][] createHisto(double xMin,double xMax,int nBins)
		{
		double[][] histo = new double[2][nBins];
		double dx = (xMax-xMin)/nBins;
		histo[0][0] = xMin + dx/2.0;

		for (int i=0; i<nBins; i++)
			{
			histo[0][i] = histo[0][0] + i*dx;
			histo[1][i] = 0.0;
			}
		return histo;
		}

	static void fillHisto(double[][] histo, double x)
		{
		int L = histo[0].length;
		double dx = (histo[0][L-1]-histo[0][0])/(L-1.0);
		double xMin = histo[0][0]-dx/2.0;
		double xMax = histo[0][L-1]+dx/2.0;
		int bin = binNb(xMin,xMax,L,x);
		if (0<=bin && bin<L) histo[1][bin]++;
		//else System.out.println("Outlier!");
		}

	static void normHisto(double[][] histo)
		{
		int L = histo[0].length;
		double sum = 0.0;
		for (int i=0; i<L; i++) sum = sum + histo[1][i];

		if (sum==0.0)
			{
			System.out.println("Histogram cannot be normalized, is empty!");
			System.exit(-9);
			}

		for (int i=0; i<L; i++) histo[1][i] = histo[1][i]/sum;
		}

	static void printHisto(double[][] histo)
		{
		System.out.println();
		int L = histo[0].length;
		for (int i=0; i<L; i++)
			{
			String xVal = String.format("%.5f", histo[0][i]-1.0);
			String yVal = String.format("%.5f", histo[1][i]);
			System.out.print( "(" + xVal + "," + yVal + ") " );
			}
		System.out.println();

		System.out.println();
		double sum = 0.0;
		for (int i=0; i<L; i++)
			{
			sum = sum + histo[0][i]*histo[1][i];
			}
		System.out.println("Histo average is = " + sum);
		}

	public static void main(String[] args)
		{
		Locale.setDefault(new Locale("US") );

		double d = 1E-4;
		double[] x = {1.0,1.0+d,1.0+d+d,1.0+d+d+d};
		double[] y = new double[4];
		for (int i=0; i<4; i++) y[i] = fn(x[i]);

		double[][] M = new double[4][4];
		for (int i=0; i<4; i++)
			{
			M[i][0] = 1.0;
			M[i][1] = cos( x[i] );
			M[i][2] = cos( 2.0*x[i] );
			M[i][3] = cos( 3.0*x[i] );
			}

		Matrix Mat = new Matrix(M);
		Matrix iMat = Mat.inverse();

		//Mat.print(4,4);
		//iMat.print(4,4);

		Matrix yVec = new Matrix(y,4);
		Matrix cfsVec_single = iMat.times( yVec );
		double[][] cfs_single = cfsVec_single.getArray();

		Random rng = new SecureRandom();

		double[] khanSum = new double[4];
		double[] khanC = new double[4];
		for (int i=0; i<4; i++)
			{
			khanSum[i]=0.0;
			khanC[i]=0.0;
			}

		double xMin = truth-width;
		double xMax = truth+width;
		double[][] histo = createHisto(xMin,xMax,nHistoBins);

		for (int i=0; i<stat; i++)
			{
			Matrix pertM = new Matrix(4,4);

			for (int k=0; k<4; k++)
				{
				for (int l=0; l<4; l++) pertM.set(k,l, Mat.get(k,l)+eps*( 2.0*rng.nextDouble()-1.0) );
				}

			Matrix iPertM = pertM.inverse();
			double[][] cPert = iPertM.times(yVec).getArray();

			double[] khanY = new double[4];
			double[] khanT = new double[4];
			for (int j=0; j<4; j++)
				{
				khanY[j] = cPert[j][0] - khanC[j];
				khanT[j] = khanSum[j] + khanY[j];
				khanC[j] = (khanT[j] - khanSum[j]) - khanY[j];
				khanSum[j] = khanT[j];
				}
			fillHisto(histo,cPert[0][0]);
			}

		System.out.println("Truth");
		for (int i=0; i<4; i++) System.out.print(cfs[i] + "\t" );
		System.out.println();

		System.out.println();
		System.out.println("AVERAGED");
		for (int i=0; i<4; i++) System.out.print(khanSum[i]/stat + "\t");
		System.out.println();
		for (int i=0; i<4; i++) System.out.print(abs(khanSum[i]/stat - cfs[i]) + "\t");
		System.out.println();

		System.out.println();
		System.out.println("SINGLE");
		for (int i=0; i<4; i++) System.out.print(cfs_single[i][0] + "\t");
		System.out.println();
		for (int i=0; i<4; i++) System.out.print(abs(cfs_single[i][0] - cfs[i]) + "\t");
		System.out.println();

		System.out.println();
		System.out.println("RATIOS");
		for (int i=0; i<4; i++) System.out.print(abs(cfs_single[i][0] - cfs[i])/abs(khanSum[i]/stat - cfs[i]) + "\t");

		System.out.println();
		System.out.println("HISTO C0");
		normHisto(histo);
		printHisto(histo);
		}

	}
